(* lit le fichier [filename] et retourne la chaîne des caractères contenus dans
le fichier *)
let string_of_file (filename: string) : string =
  let i = open_in filename in
  let rec aux (acc: char list) : string =
    match 
      try
        Some (input_char i)
      with
      | End_of_file -> None
    with
    | Some(c) -> aux (c :: acc)
    | None -> acc |> List.rev |> List.to_seq |> String.of_seq
  in
  let res = aux [] in
  close_in i ; res

(* écrit la chaîne de caractère [s] dans le fichier de nom [filename] *)
let file_of_string (filename: string) (s: string): unit =
  let o = open_out filename in
  let n = String.length s in
  let rec aux (i: int) : unit =
    if i < n then (output_char o s.[i]; aux (i+1))
  in
  aux 0; close_out o

(* Fonctions auxiliaires *)
let cst = 8

let bit_sequence_of_char (c: char): bool list =
  let ith_bit i = ((int_of_char c) land (1 lsl i)) lsr i = 1 in
  List.map ith_bit (List.init cst (fun i -> cst-1-i))

let char_of_bit_sequence (bl: bool list) : char * int =
  let padder (bl: bool list) : bool list * int =
    let n = List.length bl in
    bl @ (List.init (cst-n) (fun _ -> false)), cst-n
  in
  let int_of_bool (b: bool): int = if b then 1 else 0 in
  let bl, padded = padder bl in
  let ith_partic i = (((List.nth bl (cst-1-i)) |> int_of_bool) lsl i) in  
  let res_int = List.fold_left (fun acc i -> acc + (ith_partic i)) 0 (List.init cst (fun i -> i))
  in (res_int |> char_of_int , padded)


(* lit le fichier nommé [filename] préfixé par un octet indiquant 
  le nombre de bits significatifs dans le dernier octet. *)
let bit_sequence_of_file_with_mod (filename: string): bool list =
  let s = string_of_file filename in
  let n = String.length s in
  let last_block_size = s.[0] |> int_of_char in
  let rec aux i res =
    if i = 0 then res
    else if i = n-1 then aux (i-1) (((bit_sequence_of_char (s.[i])) |> List.filteri (fun j _ -> j < last_block_size)) @ res)
    else aux (i-1) ((bit_sequence_of_char (s.[i])) @ res)
  in aux (n-1) []

(* écrit la séquence de bits [bit_sequence] ds un fichier [filename]. 
	Le premier octet du fichier est utilisé pour coder le nombre 
	de bits significatifs du dernier octet *)
let file_with_mod_of_bit_sequence (filename: string) (bit_sequence: bool list): unit =
  let take (n: int) (l: 'a list): 'a list * 'a list = 
    let rec aux (n: int) (l: 'a list) (acc: 'a list) =
      match n, l with
      | 0, _ -> (List.rev acc), l
      | _, p::q -> aux (n-1) q (p :: acc)
      | _ -> (List.rev acc), l
    in aux n l []
  in
  let how_much = let v = (List.length bit_sequence) mod cst in if v = 0 then cst else v in 
  let o = open_out filename in
  how_much |> char_of_int |> output_char o ;
  let rec aux (bit_sequence: bool list) =
    match bit_sequence with
    | [] -> ()
    | _ ->
      begin
        let to_write, rest = take cst bit_sequence in
        let c, _ = char_of_bit_sequence to_write in
        output_char o c;
        aux rest
      end
  in aux bit_sequence; close_out o


let tab_e_from_arbre_d (a:arbre_d) : tab_e =
	(* hyp : *)
	(* calcule le tableau d'encodage associé à l'arbre a *)
	let t = Array.make 256 None in
	let rec aux (aa:arbre_d)(co:codage) :unit = 
		match aa with 
		| Feuille(c) -> t.(int_of_char c) <- Some(List.rev co)
		| Ndd (g,d) -> aux g (false::co); aux d (true::co); 
	in aux a []; t
	
	
let rec cod_from_char (t:tab_e)(c:char): codage =
	(* hyp : t a bien un codage enregistré pour c *)
	(* calcule le codage du caractère c selon t *)
	match t.(int_of_char c) with
		| None -> failwith ("erreur pour le caractère "^(String.make 1 c))
		| Some (co) -> co

